/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

/**
* Unit tests for NGS Pileup interface, CSRA1 based implementation
*/

#include "ngsfixture.hpp"

#include <kapp/args.h> /* ArgsMakeAndHandle */

#include <kfg/config.h> /* KConfigDisableUserSettings */

#include <ngs/ncbi/NGS.hpp>

#include <ktst/unit_test.hpp>

#include <sysalloc.h>
#include <assert.h>
#include <memory.h> // memset
#include <stdio.h>

#include <sstream>

using namespace ncbi::NK;

static rc_t argsHandler(int argc, char* argv[]);
TEST_SUITE_WITH_ARGS_HANDLER(NgsCsra1PileupCppTestSuite, argsHandler);

FIXTURE_TEST_CASE(CSRA1_PileupIterator_GetDepth, NgsFixture)
{
    char const db_path[] = "SRR341578";

    std::vector <uint32_t> vecDepthSlice, vecDepthEntire, vecRef;

    int64_t const pos_start = 20017;
    uint64_t const len = 5;

    vecRef.push_back(1); // 20017
    vecRef.push_back(0); // 20018
    vecRef.push_back(1); // 20019
    vecRef.push_back(1); // 20020
    vecRef.push_back(3); // 20021

    {
        ngs::ReadCollection run = open (db_path);
        ngs::ReferenceIterator ri = run.getReferences ();

        ri.nextReference ();
        ri.nextReference ();

        ngs::PileupIterator pi = ri.getPileups ( ngs::Alignment::primaryAlignment );

        uint64_t ref_pos = 0;
        for (; pi.nextPileup (); ++ ref_pos)
        {
            if ( ref_pos >= (uint64_t)pos_start && ref_pos < (uint64_t)pos_start + len )
                vecDepthEntire.push_back ( pi.getPileupDepth () );
        }
    }
    {
        ngs::ReadCollection run = open (db_path);
        ngs::ReferenceIterator ri = run.getReferences ();

        ri.nextReference ();
        ri.nextReference ();

        ngs::PileupIterator pi = ri.getPileupSlice ( pos_start, len, ngs::Alignment::primaryAlignment );

        uint64_t ref_pos = (uint64_t)pos_start;
        for (; pi.nextPileup (); ++ ref_pos)
        {
            if ( ref_pos >= (uint64_t)pos_start && ref_pos < (uint64_t)pos_start + len )
                vecDepthSlice.push_back ( pi.getPileupDepth () );
        }
    }

    REQUIRE_EQ ( vecRef.size(), vecDepthEntire.size() );
    REQUIRE_EQ ( vecRef.size(), vecDepthSlice.size() );

    for ( size_t i = 0; i < (size_t)len; ++i )
    {
        REQUIRE_EQ ( vecRef [i], vecDepthEntire [i] );
        REQUIRE_EQ ( vecRef [i], vecDepthSlice [i] );
    }
}

FIXTURE_TEST_CASE(CSRA1_PileupEventIterator_GetType, NgsFixture)
{
    char const db_path[] = "SRR341578";

    int64_t const pos_start = 20022;
    uint64_t const len = 1;

    ngs::ReadCollection run = open (db_path);
    ngs::ReferenceIterator ri = run.getReferences ();

    ri.nextReference ();
    ri.nextReference ();

    ngs::PileupEvent::PileupEventType arrRefEvents [] =
    {
        (ngs::PileupEvent::PileupEventType)(ngs::PileupEvent::mismatch | ngs::PileupEvent::alignment_minus_strand),
        ngs::PileupEvent::mismatch,
        ngs::PileupEvent::mismatch,
        (ngs::PileupEvent::PileupEventType)(ngs::PileupEvent::mismatch | ngs::PileupEvent::alignment_start),
        (ngs::PileupEvent::PileupEventType)(ngs::PileupEvent::mismatch | ngs::PileupEvent::alignment_minus_strand  | ngs::PileupEvent::alignment_start),
        (ngs::PileupEvent::PileupEventType)(ngs::PileupEvent::mismatch | ngs::PileupEvent::alignment_start)
    };

    ngs::PileupIterator pi = ri.getPileupSlice ( pos_start, len, ngs::Alignment::primaryAlignment );

    for (; pi.nextPileup (); )
    {
        REQUIRE_EQ ( pi.getPileupDepth(), (uint32_t)6 );
        for (size_t i = 0; pi.nextPileupEvent (); ++i)
        {
            REQUIRE_EQ ( pi.getEventType (), arrRefEvents [i] );
        }
    }
}

struct PileupEventStruct
{
    ngs::PileupEvent::PileupEventType event_type;
    ngs::PileupEvent::PileupEventType next_event_type;
    uint32_t repeat_count, next_repeat_count;
    int mapping_quality;
    char alignment_base;
    bool deletion_after_this_pos;
    ngs::String insertion_bases;
};

struct PileupLine
{
    typedef std::vector <PileupEventStruct> TEvents;

    uint32_t depth;
    TEvents vecEvents;
};

void print_line (
    PileupLine const& line,
    char const* name,
    int64_t pos_start,
    int64_t pos,
    ngs::String const& strRefSlice,
    std::ostream& os)
{
    os
        << name
        << "\t" << (pos + 1)    // + 1 to be like sra-pileup - 1-based position
        << "\t" << strRefSlice [pos - pos_start]
        << "\t" << line.depth
        << "\t";

    for (PileupLine::TEvents::const_iterator cit = line.vecEvents.begin(); cit != line.vecEvents.end(); ++ cit)
    {
        PileupEventStruct const& pileup_event = *cit;

        ngs::PileupEvent::PileupEventType eventType = pileup_event.event_type;

        if ( ( eventType & ngs::PileupEvent::alignment_start ) != 0 )
        {
            int32_t c = pileup_event.mapping_quality + 33;
            if ( c > '~' ) { c = '~'; }
            if ( c < 33 ) { c = 33; }

            os << "^" << (char)(c);
        }

        bool reverse = ( eventType & ngs::PileupEvent::alignment_minus_strand ) != 0;

        switch ( eventType & 7 )
        {
        case ngs::PileupEvent::match:
            os << (reverse ? "," : ".");
            break;
        case ngs::PileupEvent::mismatch:
            os
                << (reverse ?
                (char)tolower( pileup_event.alignment_base )
                : (char)toupper( pileup_event.alignment_base ));
            break;
        case ngs::PileupEvent::deletion:
            os << (reverse ? "<" : ">");
            break;
        }

        if ( pileup_event.insertion_bases.size() != 0 )
        {
            bool next_reverse = ( pileup_event.next_event_type & ngs::PileupEvent::alignment_minus_strand ) != 0;
            os
                << "+"
                << pileup_event.insertion_bases.size();

            for ( uint32_t i = 0; i < pileup_event.insertion_bases.size(); ++i )
            {
                os
                    << (next_reverse ?
                    (char)tolower(pileup_event.insertion_bases[i])
                    : (char)toupper(pileup_event.insertion_bases[i]));
            }
        }


        if ( pileup_event.deletion_after_this_pos )
        {
            uint32_t count = pileup_event.next_repeat_count;
            os << "-" << count;

            for ( uint32_t i = 0; i < count; ++i )
            {
                os
                    << (reverse ?
                    (char)tolower(strRefSlice [pos - pos_start + i + 1]) // + 1 means "deletion is at the NEXT position"
                    : (char)toupper(strRefSlice [pos - pos_start + i + 1])); // + 1 means "deletion is at the NEXT position"
            }

        }

        if ( ( eventType & ngs::PileupEvent::alignment_stop ) != 0 )
            os << "$";
    }
    os << std::endl;
}

void clear_line ( PileupLine& line )
{
    line.depth = 0;
    line.vecEvents.clear ();
}

void mark_line_as_starting_deletion ( PileupLine& line, uint32_t repeat_count, size_t alignment_index )
{
    PileupEventStruct& pileup_event = line.vecEvents [ alignment_index ];
    if ( ( pileup_event.event_type & 7 ) != ngs::PileupEvent::deletion)
    {
        pileup_event.next_repeat_count = repeat_count;
        pileup_event.deletion_after_this_pos = true;
    }
}

void mark_line_as_starting_insertion ( PileupLine& line, ngs::String const& insertion_bases, size_t alignment_index, ngs::PileupEvent::PileupEventType next_event_type )
{
    PileupEventStruct& pileup_event = line.vecEvents [ alignment_index ];
    pileup_event.insertion_bases = insertion_bases;
    pileup_event.next_event_type = next_event_type;
}

void mimic_sra_pileup (
            ngs::ReadCollection run,
            char const* ref_name,
            ngs::Alignment::AlignmentCategory category,
            int64_t const pos_start, uint64_t const len,
            std::ostream& os)
{
    ngs::Reference r = run.getReference ( ref_name );
    ngs::String const& canonical_name = r.getCanonicalName ();

    // in strRefSlice we want to have bases to report current base and deletions
    // for current base it would be enough to have only slice [pos_start, len]
    // but for deletions we might have situation when we want
    // to report a deletion that goes beyond (pos_start + len) on the reference
    // so we have to read some bases beyond our slice end
    ngs::String strRefSlice = r.getReferenceBases ( pos_start, len + 10000 );

    ngs::PileupIterator pi = r.getPileupSlice ( pos_start, len, category );

    PileupLine line_prev, line_curr;

    // maps current line alignment vector index to
    // previous line alignment vector index
    // mapAlignmentIdx[i] contains adjustment for index, not the absolute value
    std::vector <int64_t> mapAlignmentIdxPrev, mapAlignmentIdxCurr;

    int64_t pos = pos_start;
    for (; pi.nextPileup (); ++ pos)
    {
        line_curr.depth = pi.getPileupDepth ();
        line_curr.vecEvents.reserve (line_curr.depth);
        mapAlignmentIdxCurr.reserve (line_curr.depth);

        int64_t current_stop_count = 0; // number of encountered stops
        bool increased_stop_count = false; // we have increased count (skipped position) on the last step

        for (; pi.nextPileupEvent (); )
        {
            PileupEventStruct pileup_event;

            pileup_event.deletion_after_this_pos = false;
            pileup_event.event_type = pi.getEventType ();

            if ( ( pileup_event.event_type & ngs::PileupEvent::alignment_start ) != 0 )
                pileup_event.mapping_quality = pi.getMappingQuality();

            if ((pileup_event.event_type & 7) == ngs::PileupEvent::mismatch)
                pileup_event.alignment_base = pi.getAlignmentBase();

            if (increased_stop_count)
            {
                if (pileup_event.event_type & ngs::PileupEvent::alignment_stop)
                {
                    ++current_stop_count;
                    mapAlignmentIdxCurr [mapAlignmentIdxCurr.size() - 1] = current_stop_count;
                }
                else
                    increased_stop_count = false;
            }
            else
            {
                if (pileup_event.event_type & ngs::PileupEvent::alignment_stop)
                {
                    ++current_stop_count;
                    increased_stop_count = true;
                }
                mapAlignmentIdxCurr.push_back ( current_stop_count );
            }

            if ( pos != pos_start )
            {
                // here in mapAlignmentIdxPrev we have already initialized
                // indicies for line_prev
                // so we can find corresponding alignment by doing:
                // int64_t idx = line_curr.vecEvents.size()
                // line_prev.vecEvents [ idx + mapAlignmentIdxPrev [idx] ]

                size_t idx_curr_align = line_curr.vecEvents.size();

                if (mapAlignmentIdxPrev.size() > idx_curr_align)
                {
                    if ((pileup_event.event_type & 7) == ngs::PileupEvent::deletion)
                        mark_line_as_starting_deletion ( line_prev, pi.getEventRepeatCount(), mapAlignmentIdxPrev [idx_curr_align] + idx_curr_align );
                    if ( pileup_event.event_type & ngs::PileupEvent::insertion )
                        mark_line_as_starting_insertion ( line_prev, pi.getInsertionBases().toString(), mapAlignmentIdxPrev [idx_curr_align] + idx_curr_align, pileup_event.event_type );
                }
            }

            line_curr.vecEvents.push_back ( pileup_event );
        }

        if ( pos != pos_start ) // there is no line_prev for the first line - nothing to print
        {
            // print previous line
            print_line ( line_prev, canonical_name.c_str(), pos_start, pos - 1, strRefSlice, os );
        }

        line_prev = line_curr;
        mapAlignmentIdxPrev = mapAlignmentIdxCurr;

        clear_line ( line_curr );
        mapAlignmentIdxCurr.clear();
    }
    // TODO: if the last line should contain insertion or deletion start ([-+]<number><seq>)
    // we have to look ahead 1 more position to be able to discover this and
    // modify line_prev, but if the last line is the very last one for the whole
    // reference - we shouldn't do that. This all isn't implemented yet in this function
    print_line ( line_prev, canonical_name.c_str(), pos_start, pos - 1, strRefSlice, os );
}

// Moved to CSRA1PileUpTest_Slow.cpp, to be run under 'make slowtests'
//
// FIXTURE_TEST_CASE(CSRA1_PileupEventIterator_AdjacentIndels, NgsFixture)
// {
//     // This test crashed in CSRA1_PileupEvent.c because of
//     // insertion followed by deletion (see vdb-dump SRR1164787 -T PRIMARY_ALIGNMENT -R 209167)
//     // So now this test has to just run with no crashes to be considered as passed successfully

//     char const db_path[] = "SRR1164787";
//     char const ref_name[] = "chr1";

//     int64_t const pos_start = 1386093;
//     int64_t const pos_end = 9999999;
//     uint64_t const len = (uint64_t)(pos_end - pos_start + 1);

//     std::ostringstream sstream;

//     mimic_sra_pileup ( open(db_path), ref_name, ngs::Alignment::primaryAlignment, pos_start, len, sstream );
// }

FIXTURE_TEST_CASE(CSRA1_PileupEventIterator_DeletionAndEnding, NgsFixture)
{

    //sra-pileup SRR497541 -r "Contig307.Contig78.Contig363_2872688_2872915.Contig307.Contig78.Contig363_1_2872687":106436-106438 -s -n
    // There should be no "g-3taa$" for the position 106438, only "g$" must be there

    char const db_path[] = "SRR497541";
    char const ref_name[] = "Contig307.Contig78.Contig363_2872688_2872915.Contig307.Contig78.Contig363_1_2872687";

    int64_t const pos_start = 106436-1;
    int64_t const pos_end = 106438 - 1;
    uint64_t const len = (uint64_t)(pos_end - pos_start + 1);

    std::ostringstream sstream;
    std::ostringstream sstream_ref;

    sstream_ref << ref_name << "\t106436\tT\t106\tC$C$,$.,$,$,$,$.-5TATAA.-5TATAA,-5tataa.-5TATAA.-5TATAA.$.-5TATAA.-5TATAA,-5tataa.-5TATAA,-5tataa,-5tataa.-5TATAA,-5tataa,-5tataa.-5TATAA,-5tataa,-5tataa,-5tataa,-5tataa.-5TATAA.-5TATAA,-5tataa.-5TATAA,-5tataa,$.-5TATAA,-5tataa,,-5tataa,$.-5TATAA,-5tataa,-5tataa.-5TATAA,-5tataa.$.$.-5TATAA.-5TATAA.-5TATAA.-5TATAA,-5tataa.-5TATAA.-5TATAA.-5TATAA.-5TATAA.-5TATAA.-5TATAA.-5TATAA,-5tataa.-5TATAA,-5tataa,-5tataa,-5tataa.-5TATAA.$,-5tataa.-5TATAA.-5TATAA.-5TATAA.-5TATAA,-5tataa,-5tataa,-5tataa.-5TATAA,-5tataa,-5tataa.-5TATAA,-5tataa,-5tataa.-5TATAA.-5TATAA.$.-5TATAA,-5tataa.-5TATAA.-5TATAA.-5TATAA.-5TATAA,-5tataa,-5tataa.-5TATAA,-5tataa.-5TATAA.-5TATAA.-5TATAA,-5tataa,-5tataa,-5tataa,-5tataa,-5tataa,-5tataa.-5TATAA.-5TATAA,-5tataa,-5tataa^]g" << std::endl;
    sstream_ref << ref_name << "\t106437\tT\t92\tC$>><>>>><><<><<><<<<>><><><c<><<><>>>><>>>>>>><><<<><>>>><<<><<><<>>><>>>><<><>>><<<<<<>><<," << std::endl;
    sstream_ref << ref_name << "\t106438\tA\t91\t>><>>>><><<><<><<<<>><><><g$<><<><>>>><>>>>>>><><<<><>>>><<<><<><<>>><>>>><<><>>><<<<<<>><<," << std::endl;

    mimic_sra_pileup ( open(db_path), ref_name, ngs::Alignment::primaryAlignment, pos_start, len, sstream );

    REQUIRE_EQ ( sstream.str (), sstream_ref.str () );
}

FIXTURE_TEST_CASE(CSRA1_PileupEventIterator_LongDeletions, NgsFixture)
{

    //SRR1113221 "chr4" 10204 10208
    // 1. At the position 10205 we want to have the full lenght (6) deletion
    // 2. Previously we had a misplaced insertion at the position 10204 (+3att)

    char const db_path[] = "SRR1113221";

    int64_t const pos_start = 10204-1;
    int64_t const pos_end = 10208 - 1;
    uint64_t const len = (uint64_t)(pos_end - pos_start + 1);

    // The output must be the same as for "sra-pileup SRR1113221 -r "chr4":10204-10208 -s -n"
    std::ostringstream sstream;
    std::ostringstream sstream_ref;

    sstream_ref << "NC_000004.11\t10204\tT\t37\ta,$c+2tc,,+3att.$.$.+2TT.,,.......,..<......,...g...^!." << std::endl;
    sstream_ref << "NC_000004.11\t10205\tC\t37\t,$,t+3ttt,..,t.......,A.<....-6TCGGCT.-6TCGGCT.,...,....^$,^!.^$." << std::endl;
    sstream_ref << "NC_000004.11\t10206\tT\t40\t,,,$..a$g$....$.$N.,..,...-6CGGCTG>>.,...-6CGGCTGa....,..^$.^$.^%.^$." << std::endl;
    sstream_ref << "NC_000004.11\t10207\tC\t40\t,$,$.$.$.....,..,..>>>.,..>t-6ggctgc...+2TC.,......^$.^$.^$.^$.^A." << std::endl;
    sstream_ref << "NC_000004.11\t10208\tG\t39\t.....,..,..>>>.,..><....,...........^$,^M.^$." << std::endl;

    mimic_sra_pileup ( open(db_path), "chr4", ngs::Alignment::primaryAlignment, pos_start, len, sstream );

    REQUIRE_EQ ( sstream.str (), sstream_ref.str () );
}

FIXTURE_TEST_CASE(CSRA1_PileupEventIterator_Deletion, NgsFixture)
{
    // deletions
    char const db_path[] = "SRR341578";

    int64_t const pos_start = 2427;
    int64_t const pos_end = 2428;
    uint64_t const len = (uint64_t)(pos_end - pos_start + 1);

    // The output must be the same as for "sra-pileup SRR341578 -r NC_011752.1:2428-2429 -s -n"
    std::ostringstream sstream;
    std::ostringstream sstream_ref;

    sstream_ref << "NC_011752.1\t2428\tG\t34\t..,.,.-1A.-1A.-1A.-1A.-1A.-1A,-1a,-1a.-1A,-1a,-1a.-1A.-1A.-1A,-1a,-1a,-1a,-1a.-1A,-1a,-1a.-1A,-1a.-1A.-1A,-1a.^F,^F," << std::endl;
    sstream_ref << "NC_011752.1\t2429\tA\t34\t.$.$,$.,>>>>>><<><<>>><<<<><<><>><Ggg" << std::endl;

    mimic_sra_pileup ( open(db_path), "gi|218511148|ref|NC_011752.1|", ngs::Alignment::all, pos_start, len, sstream );

    REQUIRE_EQ ( sstream.str (), sstream_ref.str () );
}

FIXTURE_TEST_CASE(CSRA1_PileupEventIterator_Insertion, NgsFixture)
{
    // simple matches, mismatch, insertion, mapping quality
    char const db_path[] = "SRR341578";

    int64_t const pos_start = 2017;
    int64_t const pos_end = 2018;
    uint64_t const len = (uint64_t)(pos_end - pos_start + 1);

    // The output must be the same as for "sra-pileup SRR341578 -r NC_011752.1:2018-2019 -s -n"
    std::ostringstream sstream;
    std::ostringstream sstream_ref;

    sstream_ref << "NC_011752.1\t2018\tT\t17\t.....,,A,,..+2CA..^F.^F.^:N" << std::endl;
    sstream_ref << "NC_011752.1\t2019\tC\t19\t.....,,.,,.......^F.^F," << std::endl;

    mimic_sra_pileup ( open(db_path), "gi|218511148|ref|NC_011752.1|", ngs::Alignment::all, pos_start, len, sstream );

    REQUIRE_EQ ( sstream.str (), sstream_ref.str () );
}

FIXTURE_TEST_CASE(CSRA1_PileupEventIterator_TrickyInsertion, NgsFixture)
{
    // the insertion occurs in 1 or more previous chunks but not the current

    char const db_path[] = "SRR341578";

    int64_t const pos_start = 380000;
    int64_t const pos_end = 380001;
    uint64_t const len = (uint64_t)(pos_end - pos_start + 1);

    // The output must be the same as for "sra-pileup SRR341578 -r NC_011748.1:380001-380002 -s -n"
    std::ostringstream sstream;
    std::ostringstream sstream_ref;

    sstream_ref << "NC_011748.1\t380001\tT\t61\t....,,...,......,,...,,.....,,..,.,,,,...,,,,,,,+2tc.,.....G....," << std::endl;
    sstream_ref << "NC_011748.1\t380002\tT\t61\t.$.$.$.$,$,$...,......,,...,,.....,,A.,.,,,,...,,,,,,,.,.....G....," << std::endl;

    mimic_sra_pileup ( open(db_path), "gi|218693476|ref|NC_011748.1|", ngs::Alignment::primaryAlignment, pos_start, len, sstream );

    REQUIRE_EQ ( sstream.str (), sstream_ref.str () );
}

FIXTURE_TEST_CASE(CSRA1_PileupIterator_StartingZeros, NgsFixture)
{
    // this is transition from depth == 0 to depth == 1
    // initial code had different output for primaryAlignments vs all

    int64_t const pos_start = 19374;
    int64_t const pos_end = 19375;
    uint64_t const len = (uint64_t)(pos_end - pos_start + 1); //3906625;

    // when requesting category == all, the output must be the same as with
    // primaryAlignments
    // reference output: sra-pileup SRR833251 -r "gi|169794206|ref|NC_010410.1|":19375-19376 -s -n

    std::ostringstream sstream;
    std::ostringstream sstream_ref;

    sstream_ref << "gi|169794206|ref|NC_010410.1|\t19375\tC\t0\t" << std::endl;
    sstream_ref << "gi|169794206|ref|NC_010410.1|\t19376\tA\t1\t^!." << std::endl;

    mimic_sra_pileup ( open("SRR833251"), "gi|169794206|ref|NC_010410.1|", ngs::Alignment::all, pos_start, len, sstream );

    REQUIRE_EQ ( sstream.str (), sstream_ref.str () );
}

FIXTURE_TEST_CASE(CSRA1_PileupIterator_MapQuality, NgsFixture)
{
    // different mapping quality
    // there was a bug caused by usage of char c instead int32_t c
    // in printing code inside mimic_sra_pileup

    int64_t const pos_start = 183830-1;
    int64_t const pos_end = pos_start;
    uint64_t const len = (uint64_t)(pos_end - pos_start + 1);

    std::ostringstream sstream;
    std::ostringstream sstream_ref;

    sstream_ref << "gi|169794206|ref|NC_010410.1|\t183830\tA\t1\t^~," << std::endl;

    mimic_sra_pileup ( open("SRR833251"), "gi|169794206|ref|NC_010410.1|", ngs::Alignment::all, pos_start, len, sstream );

    REQUIRE_EQ ( sstream.str (), sstream_ref.str () );
}

FIXTURE_TEST_CASE(CSRA1_PileupIterator_Depth, NgsFixture)
{
    // if ngs::Alignment::all is used here
    // there will be discrepancy with sra-pileup

    int64_t const pos_start = 519533-1;
    int64_t const pos_end = pos_start;
    uint64_t const len = (uint64_t)(pos_end - pos_start + 1);

    std::ostringstream sstream;
    std::ostringstream sstream_ref;

    sstream_ref << "gi|169794206|ref|NC_010410.1|\t519533\tC\t1\t," << std::endl;

    mimic_sra_pileup ( open("SRR833251"), "gi|169794206|ref|NC_010410.1|", ngs::Alignment::primaryAlignment, pos_start, len, sstream );

    REQUIRE_EQ ( sstream.str (), sstream_ref.str () );
}

FIXTURE_TEST_CASE(CSRA1_PileupIterator_TrailingInsertion, NgsFixture)
{
    // Loaders sometimes fail and produce a run with trailing insertions
    // Like now (2015-06-29) SRR1652532 for SRR1652532.PA.97028

    int64_t const pos_start = 42406728-1;
    int64_t const pos_end = 42406732;
    uint64_t const len = (uint64_t)(pos_end - pos_start + 1);

    // reference output: sra-pileup SRR1652532 -r "CM000671.1":42406728-42406732 -s -n

    // as of 2015-06-25, this command causes sra-pileup to crash, so
    // we don't compare our output with one of sra-pileup
    // and only checking that this test doesn't crash

    std::ostringstream sstream;
    //std::ostringstream sstream_ref;

    //sstream_ref << "gi|169794206|ref|NC_010410.1|\t19375\tC\t0\t" << std::endl;
    //sstream_ref << "gi|169794206|ref|NC_010410.1|\t19376\tA\t1\t^!." << std::endl;

    mimic_sra_pileup ( open("SRR1652532"), "CM000671.1", ngs::Alignment::all, pos_start, len, sstream );

    //std::cout << sstream.str() << std::endl;

    //REQUIRE_EQ ( sstream.str (), sstream_ref.str () );
}

#if 0 /* TODO: this test needs to be investigated later */
FIXTURE_TEST_CASE(CSRA1_PileupIterator_FalseMismatch, NgsFixture)
{
    // here is two problems:
    // 1. at the position 19726231 GetEventType returns "mismatch"
    //    but the base == 'C' for both reference and alignment
    // 2. ngs code doesn't report deletion in the beginning of both lines
    //    and also it reports depths: 1 and 3, while sra-pileup returns 2 and 4
    // And also it produces a lot of warnings in stderr in the debug version

    // Update:
    // (1) reproduced in pileup-stats -v -a primary -x 2 -e 2 SRR1652532 -o SRR1652532.out
    //     due to the bug in new pileup code - need to start not from position 19726231
    //     but ~25 before that
    // (2). still remains
    // 3. if one uncomments REQUIRE_EQ in the end a lot of debug warning appear in the stderr

    int64_t const pos_start = 19726231-1;
    int64_t const pos_end = pos_start + 1;
    uint64_t const len = (uint64_t)(pos_end - pos_start + 1);

    std::ostringstream sstream;
    std::ostringstream sstream_ref;

    // sra-pileup SRR1652532 -r CM000663.1:19726231-19726232 -s -n -t p

    sstream_ref << "CM000663.1\t19726231\tG\t2\t<." << std::endl;
    sstream_ref << "CM000663.1\t19726232\tC\t4\t<.^:,^H," << std::endl;

    mimic_sra_pileup ( "SRR1652532", "CM000663.1", ngs::Alignment::primaryAlignment, pos_start, len, sstream );

    //std::cout << sstream.str() << std::endl;

    //REQUIRE_EQ ( sstream.str (), sstream_ref.str () );
}
#endif

//////////////////////////////////////////// Main

static rc_t argsHandler(int argc, char* argv[]) {
    return ArgsMakeAndHandle ( NULL, argc, argv, 0, NULL, 0 );
}

extern "C"
{

#include <kapp/args.h>

ver_t CC KAppVersion ( void )
{
    return 0x1000000;
}
rc_t CC UsageSummary (const char * progname)
{
    return 0;
}

rc_t CC Usage ( const Args * args )
{
    return 0;
}

const char UsageDefaultName[] = "test-ngs_csra1pileup-c++";

rc_t CC KMain ( int argc, char *argv [] )
{
    KConfigDisableUserSettings();
    rc_t rc=NgsCsra1PileupCppTestSuite(argc, argv);

    NgsFixture::ReleaseCache();
    return rc;
}

}

