/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#include <kfc/callstk.hpp>
#include <kfc/except.hpp>
#include <kfc/rsrc.hpp>

#ifndef _hpp_vdb3_kfc_except_
#define THROW( xc, fmt, ... ) \
    throw fmt
#define CONST_THROW( xc, msg ) \
    throw msg
#endif

namespace vdb3
{

    __thread CallStk const * callstk;

    CallStk :: CallStk ( const SrcLoc & sloc )
        : caller ( callstk )
        , depth ( 0 )
        , src ( sloc )
        , func ( "" )
    {
        if ( callstk != 0 )
        {
#ifdef _hpp_vdb3_kfc_rsrc_
            assert ( rsrc != 0 );
            FUNC_ENTRY ();
#endif
            CONST_THROW ( xc_program_state_violation, "not at top of call stack" );
        }

        callstk = this;
    }
}
