#include <klib/strings.h> /* SDL_CGI */

#define RESOLVER_CGI     "https://trace.ncbi.nlm.nih.gov/Traces/names/names.fcgi"
#define RESOLVER_CGI_HTTP "http://trace.ncbi.nlm.nih.gov/Traces/names/names.fcgi"
