/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _h_screen_
#define _h_screen_

#ifndef _h_tui_extern_
#include <tui/extern.h>
#endif

#ifndef _h_klib_defs_
#include <klib/defs.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

#include <tui/tui.h>

#define CHR_PER_LINE 1000

typedef struct tui_line
{
    char chars[ CHR_PER_LINE ];
    tui_ac ac[ CHR_PER_LINE ];  /* attrib and color */
    bool dirty[ CHR_PER_LINE ];
    bool line_dirty;
} tui_line;

#define LINES_PER_SCREEN 200

typedef struct tui_screen
{
    tui_line lines[ LINES_PER_SCREEN ];
    bool dirty;
} tui_screen;


void CC clr_tui_screen( tui_screen * screen, const tui_ac * v, bool dirty );

void CC print_into_screen( tui_screen * screen, int x, int y, const tui_ac * v, const char * s, uint32_t l );

void CC paint_into_screen( tui_screen * screen, const tui_rect * r, const tui_ac * v, const char c );

void CC send_tui_screen( tui_screen * screen, int n_lines, int n_cols );

#ifdef __cplusplus
}
#endif

#endif /* _h_screen */
