/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _h_kdb_table_priv_
#define _h_kdb_table_priv_

#ifndef _h_kdb_table_
#include <kdb/table.h>
#endif

#ifndef _h_klib_refcount_
#include <klib/refcount.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif


/*--------------------------------------------------------------------------
 * forwards
 */
struct KDatabase;
struct KDBManager;
struct KDirectory;


/*--------------------------------------------------------------------------
 * KTable
 *  represents a table
 *  normally implemented as a directory
 *  but may be a single archive file
 *  in either executable or streamable format
 */
struct KTable
{
    struct KDirectory const *dir;
    struct KDBManager const *mgr;
    struct KDatabase const *db;
    KRefcount refcount;
    uint8_t prerelease;
    char path [ 1 ];
};

/* Attach
 * Sever
 *  like KTableRelease, except called internally
 *  indicates that a child object is letting go...
 */
KTable *KTableAttach ( const KTable *self );
rc_t KTableSever ( const KTable *self );

void KTableGetName(KTable const *self, char const **rslt);

#ifdef __cplusplus
}
#endif

#endif /* _h_kdb_table_priv_ */
