/* py_ngs_defs.h */

#ifndef _h_ngs_defs_
#define _h_ngs_defs_

#define PY_RES_TYPE   int
#define PY_RES_OK     0
#define PY_RES_ERROR  1

#ifndef _h_klib_extern_
#include <klib/extern.h>
#endif

#endif

