/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/
#include <vdb/extern.h>

#include <sra/sradb.h>
#include <vdb/xform.h>
#include <klib/data-buffer.h>
#include <klib/text.h>
#include <klib/rc.h>
#include <sysalloc.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "name-tokenizer.h"

static
rc_t CC tokenize_spot_name_454 ( void *self, const VXformInfo *info, int64_t row_id,
                              VRowResult *rslt, uint32_t argc, const VRowData argv [] )
{
    rc_t rc;
    const char *name;
    unsigned region_start;
    unsigned q_start;
    unsigned i;
    spot_name_token_t *spot_name_tok;
    const int EXPECTED_NUMBER_OF_TOKENS = 2;
    
    assert(rslt->elem_bits == sizeof(spot_name_tok[0]) * 8);
    rslt->data->elem_bits = sizeof(spot_name_tok[0]) * 8;
    if( (rc = KDataBufferResize(rslt->data, EXPECTED_NUMBER_OF_TOKENS)) != 0 ) {
        return rc;
    }
    
    spot_name_tok = rslt->data->base;
    
    spot_name_tok[0].s.token_type = nt_unrecognized;
    spot_name_tok[0].s.position = 0;
    spot_name_tok[0].s.length = argv[0].u.data.elem_count;
    rslt->elem_count = 1;
    
    /* /^(.{7})([0-9]{2})([A-Z0-9]{5})$/ = (name, region, XY_q) = ($1, $2, $3); */
    if (argv[0].u.data.elem_count != 14 ) {
        return 0;
    }
    
    name = &((const char *)argv[0].u.data.base)[argv[0].u.data.first_elem];
    q_start = argv[0].u.data.elem_count - 5;
    region_start = q_start - 2;
    
    /* validate that XY_q =~ /[A-Z0-9]{5}/ */
    for(i = 0; i != 5; i++) {
        if(!((name[i + q_start] >= 'A' && name[i + q_start] <= 'Z') || (name[i + q_start] >= '0' && name[i + q_start] <= '9'))) {
            return 0;
        }
    }
    /* validate that region =~ /[0-9]{2}/ */
    if( !(name[region_start] >= '0' && name[region_start] <= '9') ||
        !(name[region_start + 1] >= '0' && name[region_start + 1] <= '9') ) {
        return 0;
    }
    spot_name_tok[0].s.token_type = nt_T;
    spot_name_tok[0].s.position = region_start;
    spot_name_tok[0].s.length = 2;
    
    spot_name_tok[1].s.token_type = nt_Q;
    spot_name_tok[1].s.position = q_start;
    spot_name_tok[1].s.length = 5;

    rslt->elem_count = EXPECTED_NUMBER_OF_TOKENS;
    return 0;
}

/* tokenize_spot_name
 *  scans name on input
 *  tokenizes into parts

typedef U16 NCBI:SRA:spot_name_token [ 3 ];
const U16 NCBI:SRA:name_token:unrecognized = 1;
const U16 NCBI:SRA:name_token:X            = 2;
const U16 NCBI:SRA:name_token:Y            = 3;
const U16 NCBI:SRA:name_token:Q            = 4;
const U16 NCBI:SRA:name_token:T            = 5;
const U16 NCBI:SRA:name_token:L            = 6;

extern function NCBI:SRA:spot_name_token
    NCBI:SRA:_454_:tokenize_spot_name #1 ( ascii name );
 */
VTRANSFACT_IMPL ( NCBI_SRA__454__tokenize_spot_name, 1, 0, 0 ) ( const void *self,
                  const VXfactInfo *info, VFuncDesc *rslt, const VFactoryParams *cp, const VFunctionParams *dp )
{
    rslt->variant = vftRow;
    rslt->u.rf = tokenize_spot_name_454;
    
    return 0;
}
