/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _hpp_vdb3_kfc_syserr_
#define _hpp_vdb3_kfc_syserr_

#ifndef _hpp_vdb3_kfc_string_
#include <kfc/string.hpp>
#endif

namespace vdb3
{

    /*------------------------------------------------------------------
     * forwards
     */


    /*------------------------------------------------------------------
     * functions
     */

    // returns a system error string
    String SysErr( int errno );

    // throws a runtime error
    void ThrowOSErr( U32 lineno, const String & func_name, int err_code )
        __attribute__ ( ( noreturn ) );

#define THROW_OSERR( func_name, err_code ) \
    ThrowOSErr( __LINE__, CONST_STRING ( # func_name ), err_code )
}

#endif // _hpp_vdb3_kfc_syserr_
