/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _h_sra_pacbio_
#define _h_sra_pacbio_

#ifndef _h_klib_defs_
#include <klib/defs.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif


/*==========================================================================
 * PACBIO types, constants
 */


/* hole status for pacbio spots
 */
#define sra_read_filter_t "INSDC:SRA:read_filter"
typedef uint8_t PACBIO_HOLE_STATUS;
enum
{
    SRA_PACBIO_HOLE_SEQUENCING = 0,
    SRA_PACBIO_HOLE_ANTIHOLE = 1,
    SRA_PACBIO_HOLE_FIDUCIAL = 2,
    SRA_PACBIO_HOLE_SUSPECT = 3,
    SRA_PACBIO_HOLE_ANTIMIRROR = 4,
    SRA_PACBIO_HOLE_FDZMW = 5,
    SRA_PACBIO_HOLE_FBZMW = 6,
    SRA_PACBIO_HOLE_ANTIBEAMLET = 7,
    SRA_PACBIO_HOLE_OUTSIDEFOV = 8
};


#ifdef __cplusplus
}
#endif

#endif /*  _h_sra_pacbio_ */
