/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _hpp_vdb3_kfc_plogger_
#define _hpp_vdb3_kfc_plogger_

#ifndef _hpp_vdb3_kfc_refcount_
#include <kfc/refcount.hpp>
#endif

#ifndef _hpp_vdb3_kfc_log_
#include <kfc/log.hpp>
#endif

#ifndef _hpp_vdb3_kfc_string_
#include <kfc/string.hpp>
#endif

namespace vdb3
{

    /*------------------------------------------------------------------
     * plogger_t
     *  simulates syslog-style output
     */
    class plogger_t : public Refcount
        , implements LoggerItf
    {
    public:

        // make the primoridal logger with argv[0]
        static Log make ( const char * ident );

        // format the output a la syslog
        virtual void msg ( LogLevel priority, const char * fmt, va_list args );

        ~ plogger_t ();

    private:

        plogger_t ( const String & hostproc, const String & newline );

        String hostproc;
        String newline;
    };

}

#endif // _hpp_vdb3_kfc_plogger_
