/*===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 */

#ifndef _xgapf_h_
#define _xgapf_h_

#include <xfs/xfs-defs.h>

#ifdef __cplusplus 
extern "C" {
#endif /* __cplusplus */

/*))))
 ((((   There is some primitives which allow to access dbGaP files
  ))))  
 ((((*/

/*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*
 * LYRICS:
 *
 * That file contains set of structures which will implement FUSE/DOKAN
 * access to dbGaP data files.
 * 
 * PRIMITEVES:
 *
 * XFSGapFiles - use this structure to access dbGaP data files
 *
 *_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*/

/*))
 //     Forwards
((*/
struct XFSGapFiles;

/*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*/
/*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*/
/*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*/

/*))
 //     XFSGapFiles
((*/

    /* These should be called before/after using this interface
     */
XFS_EXTERN rc_t CC XFSGapFilesInit ();
XFS_EXTERN rc_t CC XFSGapFilesDipose ();

/*))
 //     Opening only for reading
((*/
XFS_EXTERN rc_t CC XFSGapFilesOpenAccession (
                            const struct KFile ** File,
                            uint32_t ProjectId,
                            const char * Accession
                            );

XFS_EXTERN rc_t CC XFSGapFilesOpenObjectId (
                            const struct KFile ** File,
                            uint32_t ProjectId,
                            uint32_t ObjectId
                            );

/*))
 //     Attributes if possible
((*/
XFS_EXTERN rc_t CC XFSGapFilesAccessionSize (
                            uint64_t * Size,
                            uint32_t ProjectId,
                            const char * Accession
                            );

XFS_EXTERN rc_t CC XFSGapFilesAccessionDate (
                            KTime_t * Time,
                            uint32_t ProjectId,
                            const char * Accession
                            );

XFS_EXTERN rc_t CC XFSGapFilesObjectIdSize (
                            uint64_t * Size,
                            uint32_t ProjectId,
                            uint32_t ObjectId
                            );

XFS_EXTERN rc_t CC XFSGapFilesObjectIdDate (
                            KTime_t * Time,
                            uint32_t ProjectId,
                            uint32_t ObjectId
                            );

/*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*/
/*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*/
/*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*/
/*))
 //     XFSGapKart_Temp
((*/

/*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*_*/

#ifdef __cplusplus 
}
#endif /* __cplusplus */

#endif /* _xgapf_h_ */
