/*===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 */
#include <krypto/extern.h>
#include <klib/defs.h>

#include <krypto/cipher.h>
#include "cipher-priv.h"

#include <klib/rc.h>


rc_t KCipherMakeSub (KCipher ** new_cipher, kcipher_type type, uint32_t subtype)
{
    rc_t rc;

    if (new_cipher == NULL)
        rc = RC (rcKrypto, rcCipher, rcConstructing, rcSelf, rcNull);

    else
    {
        switch (type)
        {
        case kcipher_AES:
            switch (subtype)
            {
            default:
                rc = RC (rcKrypto, rcCipher, rcConstructing, rcBlockCipher, rcInvalid);
                break;

            case kcipher_AES_core:
                rc = KCipherByteMake (new_cipher, type);
                break;

            case kcipher_AES_ncbi_sim:
                rc = KCipherVecSimMake (new_cipher, type);
                break;

            case kcipher_AES_ncbi_vec:
                rc = KCipherVecRegMake (new_cipher, type);
                break;

            case kcipher_AES_ncbi_ni:
                rc = KCipherVecAesNiMake (new_cipher, type);
                break;
            }
            break;
        default:
            rc = RC (rcKrypto, rcCipher, rcConstructing, rcBlockCipher, rcInvalid);
            break;
        }
    }
    return rc;
}




/* EOF */
