/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#include <klib/extern.h>
#include <sysalloc.h>
#include <stdint.h>

LIB_EXPORT const uint32_t iso8859_1 [ 128 ] =
{
    /*  +       0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15 */
    /* 128 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    /* 144 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    /* 160 */ 160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
    /* 176 */ 176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,
    /* 192 */ 192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
    /* 208 */ 208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
    /* 224 */ 224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
    /* 240 */ 240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255
};
