/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _h_ncbi_vdb_blast_priv_
#define _h_ncbi_vdb_blast_priv_

#ifndef _h_vdb_extern_
#include <vdb/extern.h>
#endif

#include <stddef.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif


struct VdbBlast2naReader;

VDB_EXTERN uint64_t CC _VdbBlast2naReaderRead(
    const struct VdbBlast2naReader *self,
    uint32_t *status, uint64_t *read_id, size_t *starting_base,
    uint8_t *buffer, size_t buffer_size);

#ifdef __cplusplus
}
#endif

#endif /* _h_ncbi_vdb_blast_priv_ */
