/*===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 */

#include <krypto/extern.h>

#include "ncbi-priv.h"
#include "cipher-priv.h"
/* #include "blockcipher-priv.h" */

#include <klib/rc.h>

/* ----------------------------------------------------------------------
 * Make
 *
 * Create a new Null Block Cipher object.
 * The processor is checked to see if this particular version is supported on
 * this particular CPU.
 */
struct KBlockCipher;
rc_t NULLBCMEMBER(Make) (struct KBlockCipher ** new_obj)
{
    rc_t rc;

    /* Check parameter first */
    if (new_obj == NULL)
        rc = RC (rcKrypto, rcBlockCipher, rcConstructing, rcSelf, rcNull);

    else
    {
        *new_obj = NULL;

        rc = RC (rcKrypto, rcBlockCipher, rcConstructing, rcFunction, rcUnsupported);
    }
    return rc;
}

/* end of file lib/krypto/aes-ncbi.c */
