/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

#ifndef _h_klib_guid_
#define _h_klib_guid_

#ifndef _h_klib_extern_
#include <klib/extern.h>
#endif

#ifndef _h_klib_defs_
#include <klib/defs.h>
#endif

#ifndef _h_klib_rc_
#include <klib/rc.h>
#endif

#ifndef _h_klib_text_
#include <klib/text.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

/**
 * KGUIDMake
 * generate a GUID and place it in the provided String buffer.
 *
 * The format is version 4 variant 1, according to
 * [https://en.wikipedia.org/wiki/Universally_unique_identifier#Implementations]
 *
 *  xxxxxxxx-xxxx-4xxx-Nxxx-xxxxxxxxxxxx
 *
 * where
 *   x represents a randomly generated hexadecimal digit from 0 to f
 *   the 2 most significant bits of N are 10, the other 2 bits are randomly generated
 *
 * "buf"/"buf_size" [ OUT ] The output buffer, must be at least 37 characters long (includes 1 byte for 0-terminator)
 */

KLIB_EXTERN rc_t CC KGUIDMake( char * buf, size_t bufSize );

#ifdef __cplusplus
}
#endif

#endif /* _h_klib_guid_ */
