#!/bin/bash
# ===========================================================================
#
#                            PUBLIC DOMAIN NOTICE
#               National Center for Biotechnology Information
#
#  This software/database is a "United States Government Work" under the
#  terms of the United States Copyright Act.  It was written as part of
#  the author's official duties as a United States Government employee and
#  thus cannot be copyrighted.  This software/database is freely available
#  to the public for use. The National Library of Medicine and the U.S.
#  Government have not placed any restriction on its use or reproduction.
#
#  Although all reasonable efforts have been taken to ensure the accuracy
#  and reliability of the software and data, the NLM and the U.S.
#  Government do not and cannot warrant the performance or results that
#  may be obtained by using this software or data. The NLM and the U.S.
#  Government disclaim all warranties, express or implied, including
#  warranties of performance, merchantability or fitness for any particular
#  purpose.
#
#  Please cite the author in any work or product based on this material.
#
# ===========================================================================

OBJ_FILES="$1"
LIB_FILES="$2"

for f in $OBJ_FILES $LIB_FILES
do
    if [ ! -f $f.md5 ]
    then
        echo "$f.md5 not found" > /dev/stderr
        exit 5
    elif [ ! -r $f.md5 ]
    then
        echo "$f.md5 not readable" > /dev/stderr
        exit 5
    else
        cat $f.md5
    fi
done
